import numpy

from six import BytesIO

from smqtk_dataprovider import DataElement
from smqtk_image_io.interfaces.image_reader import ImageReader
from smqtk_image_io.utils.image import crop_in_bounds
from smqtk_image_io.bbox import AxisAlignedBoundingBox

from typing import Optional, Set, Dict

try:
    import PIL.Image
    IMPORT_SUCCESS = True
except ImportError:
    IMPORT_SUCCESS = False


class PilImageReader (ImageReader):
    """
    Image reader that uses PIL to load the image.

    This implementation may additionally raise an ``IOError`` when failing to
    to load an image.
    """

    @classmethod
    def is_usable(cls) -> bool:
        return IMPORT_SUCCESS

    def __init__(self, explicit_mode: Optional[str] = None) -> None:
        """
        Image reader that uses PIL to load the image.

        This reader returns an image matrix in [height, width] or [height,
        width, channel] dimension format, depending on source or explicit
        image mode requested.

        :param None|str explicit_mode:
            Optional explicit PIL Image mode to convert the image to before
            converting into a matrix. This should be one of the `PIL concept
            modes`_.

        :raises ValueError: The provided ``explicit_mode`` was not a valid mode
            string according to ``PIL.Image.MODES``.

        .. _PIL concept modes: https://pillow.readthedocs.io/en/3.1.x/handbook/
           concepts.html#concept-modes

        """
        super(PilImageReader, self).__init__()
        if explicit_mode and explicit_mode not in PIL.Image.MODES:
            raise ValueError("Given explicit image mode was not a valid mode "
                             "as reported by ``PIL.Image.MODES``. Given '{}'. "
                             "Available modes: {}."
                             .format(explicit_mode, PIL.Image.MODES))
        self._explicit_mode = explicit_mode

    def get_config(self) -> Dict:
        return {
            'explicit_mode': self._explicit_mode,
        }

    def valid_content_types(self) -> Set:
        # Explicitly load standard file format drivers.
        # - confirmed idempotent from at least version 5.3.0
        # TODO: get access to lazy-loaded file format driver extensions?
        PIL.Image.preinit()
        return set(PIL.Image.MIME.values())

    def _load_as_matrix(
        self, data_element: DataElement,
            pixel_crop: Optional[AxisAlignedBoundingBox] = None) \
            -> numpy.ndarray:
        """
        Internal method to be implemented that attempts loading an image
        from the given data element and returning it as an image matrix.

        Pre-conditions:
            - ``pixel_crop`` has a non-zero volume and is composed of integer
              types.

        :param smqtk.representation.DataElement data_element:
            DataElement to load image data from.
        :param None|smqtk.representation.AxisAlignedBoundingBox pixel_crop:
            Optional pixel crop region to load from the given data.  If this
            is provided it must represent a valid sub-region within the loaded
            image, otherwise a RuntimeError is raised.

        :raises RuntimeError: A crop region was specified but did not specify a
            valid sub-region of the image.

        :return: Numpy ndarray of the image data. Specific return format is
            implementation dependant.
        :rtype: numpy.ndarray

        """
        # We may have to add a mode where we use write_temp and load from that
        # if loading large images straight from bytes-in-memory is a problem
        # and that approach actually alleviates anything.

        # Catch and raise alternate IOError exception for readability.
        try:
            #: :type: PIL.Image.Image
            img = PIL.Image.open(BytesIO(data_element.get_bytes()))
        except IOError as ex:
            ex_str = str(ex)
            if 'cannot identify image file' in ex_str:
                raise IOError("Failed to identify image from bytes provided "
                              "by {}".format(data_element))
            else:
                # pass through other exceptions
                raise

        if pixel_crop:
            if not crop_in_bounds(pixel_crop, *img.size):
                raise RuntimeError("Crop provided not within input image. "
                                   "Image shape: {}, crop: {}"
                                   .format(img.size, pixel_crop))
            img = img.crop(pixel_crop.min_vertex.tolist() +
                           pixel_crop.max_vertex.tolist())

        # If the loaded image is not already the optionally provided
        # explicit mode, convert it.
        if self._explicit_mode and img.mode != self._explicit_mode:
            img = img.convert(mode=self._explicit_mode)

        # noinspection PyTypeChecker
        return numpy.asarray(img)



import io
import logging
from typing import Generator, Tuple, Iterable, Optional

import numpy
import PIL.Image
import PIL.ImageEnhance

from six.moves import range

from smqtk_dataprovider import DataElement
from smqtk_image_io.bbox import AxisAlignedBoundingBox


def is_loadable_image(data_element: DataElement) -> bool:
    """
    Determine if an image is able to be loaded by PIL.

    :param data_element: A data element to check
    :type data_element: DataElement

    :return: Whether or not the image is loadable
    :rtype: bool

    Example:
    >>>

    """
    log = logging.getLogger(__name__)

    try:
        PIL.Image.open(io.BytesIO(data_element.get_bytes()))
        return True
    except IOError as ex:
        # noinspection PyProtectedMember
        log.debug("Failed to convert '%s' bytes into an image "
                  "(error: %s). Skipping", data_element, str(ex))
        return False


def is_valid_element(
    data_element: DataElement,
        valid_content_types: Optional[Iterable] = None,
        check_image: bool = False) -> bool:
    """
    Determines if a given data element is valid.

    :param data_element: Data element
    :type data_element: DataElement

    :param valid_content_types: List of valid content types, or None to skip
        content type checking.
    :type valid_content_types: iterable | None

    :param check_image: Whether or not to try loading the image with PIL. This
        often catches issues that content type can't, such as corrupt images.
    :type check_image: bool

    :return: Whether or not the data element is valid
    :rtype: bool

    """
    log = logging.getLogger(__name__)

    if (valid_content_types is not None and
            data_element.content_type() not in valid_content_types):
        log.debug("Skipping file (invalid content) type for "
                  "descriptor generator (data_element='%s', ct=%s)",
                  data_element, data_element.content_type())
        return False

    if check_image and not is_loadable_image(data_element):
        return False

    return isinstance(data_element, DataElement)


def image_crop_center_levels(
    image: PIL.Image.Image,
    n_crops: int
) -> Generator[Tuple[int, PIL.Image.Image], None, None]:
    """
    Crop out one or more increasing smaller images from a base image by cutting
    off increasingly larger portions of the outside perimeter. Cropped image
    dimensions determined by the dimensions of the base image and the number of
    crops to generate.

    :param image: The base image array. This is not modified,.
    :param n_crops: Number of crops to generate.

    :return: Generator yielding paired level number and PIL.Image.Image tuples.
        Cropped images have not loaded/copied yet, so changes to the original
        image will affect them.

    """
    n_crops = int(n_crops)
    if n_crops <= 0:
        raise ValueError("Can't produce 0 or negative crops")

    #: :type: numpy.ndarray
    x_points = numpy.linspace(0, image.width, 2 + n_crops * 2, dtype=int)
    #: :type: numpy.ndarray
    y_points = numpy.linspace(0, image.height, 2 + n_crops * 2, dtype=int)

    # Outside edges of generated points in the original image size
    for i in range(1, n_crops + 1):
        # crop wants: [xmin, ymin, xmax, ymax]
        t = list(zip(x_points[[i, -i - 1]], y_points[[i, -i - 1]]))
        yield i, image.crop(t[0] + t[1])


def image_crop_quadrant_pyramid(
    image: PIL.Image.Image, n_levels: int) \
        -> Generator[Tuple[int, Tuple[int, int], PIL.Image.Image], None, None]:
    """
    Generate a number of crops based on a number of quadrant sub-partitions
    made based on the given number of levels.

    For example, 1 level would yield 4 crops from the 2x2 partition of the
    image. 2 levels would yield 20 crops for the 2x2 partition and the 4x4
    partition. 3 levels would yield the partitions of 2x2, 4x4 and 8x8 yielding
    84 crops, etc. General rule: (2^i)^2 for i in [1 n_levels]

    :param image: Image to crop in quadrant partitions
    :type image: PIL.Image.Image

    :param n_levels: Number of quadrant levels to generate crops for.
    :type n_levels: int

    :return: Generator yielding paired level, quadrant position and
        PIL.Image.Image tuples. Quadrant position is in (x, y) format.
        Crop images have not loaded/copied yet, so changes to the original
        image will affect them.
    :rtype: __generator[(int, (int, int), PIL.Image.Image)]

    """
    n_crops = int(n_levels)
    if n_crops <= 0:
        raise ValueError("Can't produce 0 or negative levels")

    for level in range(1, n_levels + 1):
        l_sq = 2**level
        xs = numpy.linspace(0, image.width, l_sq + 1,
                            endpoint=True, dtype=int)
        ys = numpy.linspace(0, image.height, l_sq + 1,
                            endpoint=True, dtype=int)
        for j in range(l_sq):
            for i in range(l_sq):
                yield (
                    level,
                    (i, j),
                    image.crop([xs[i], ys[j], xs[i + 1], ys[j + 1]])
                )


def image_crop_tiles(
    image: PIL.Image.Image, tile_width: int, tile_height: int,
        stride: Optional[Tuple[int, int]] = None) \
            -> Generator[Tuple[int, int, PIL.Image.Image], None, None]:
    """
    Crop out tile windows from the base image that have the width and height
    specified.

    If the image width or height is not evenly divisible by the tile width or
    height, respectively, then the crop out as many tiles as neatly fit
    starting from the axis origin. The remaining pixels are ignored.

    :param image: Image to crop tiles from.
    :type image: PIL.Image.Image

    :param tile_width: Tile crop width in pixels.
    :type tile_width: int

    :param tile_height: Tile crop height in pixels.
    :type tile_height: int

    :param stride: Optional tuple of integer pixel stride for cropping out sub-
        images. When this is None, the stride is the same as the width and
        height of the requested sub-images.
    :type stride: None | (int, int)

    :return: Generator yielding tuples containing a cropped image and its
        upper-left xy position in the original image.
    :rtype: __generator[(int, int, PIL.Image.Image)]

    """
    if stride:
        stride_x, stride_y = map(int, stride)
    else:
        stride_x = tile_width
        stride_y = tile_height

    # upper-left xy pixel coordinates for sub-images.
    y = 0
    while (y + tile_height) < image.height:
        x = 0
        while (x + tile_width) < image.width:
            t = image.crop([x, y, x+tile_width, y+tile_height])
            yield (x, y, t)
            x += stride_x
        y += stride_y


def image_brightness_intervals(image: PIL.Image.Image, n: int) \
        -> Generator[Tuple[int, PIL.Image.Image], None, None]:
    """
    Generate a number of images with different brightness levels using linear
    interpolation to choose levels between 0 (black) and 1 (original image) as
    well as between 1 and 2.

    Results will not include contrast level 0, 1 or 2 images.

    """
    n = int(n)
    if n <= 0:
        raise ValueError("Can't produce 0 intervals")

    b = numpy.linspace(0, 1, n+2, endpoint=True, dtype=float)
    for v in b[1:-1]:
        yield v, PIL.ImageEnhance.Brightness(image).enhance(v)
    b = numpy.linspace(1, 2, n + 2, endpoint=True, dtype=float)
    for v in b[1:-1]:
        yield v, PIL.ImageEnhance.Brightness(image).enhance(v)


def image_contrast_intervals(image: PIL.Image.Image, n: int) \
        -> Generator[Tuple[int, PIL.Image.Image], None, None]:
    """
    Generate a number of images with different contrast levels using linear
    interpolation to choose levels between 0 (black) and 1 (original image) as
    well as between 1 and 2.

    Results will not include contrast level 0, 1 or 2 images.

    """
    n = int(n)
    if n <= 0:
        raise ValueError("Can't produce 0 intervals")

    b = numpy.linspace(0, 1, n + 2, endpoint=True, dtype=float)
    for v in b[1:-1]:
        yield v, PIL.ImageEnhance.Contrast(image).enhance(v)
    b = numpy.linspace(1, 2, n + 2, endpoint=True, dtype=float)
    for v in b[1:-1]:
        yield v, PIL.ImageEnhance.Contrast(image).enhance(v)


def crop_in_bounds(
    bbox: AxisAlignedBoundingBox, im_width: int,
        im_height: int) -> bool:
    """
    Check if this crop specification is within a given parent bounds
    specification.

    Thus function does NOT care if the input bounding box is integer aligned.

    :param smqtk.representation.AxisAlignedBoundingBox bbox:
        Bounding box representing a sub-region within an image. This must be a
        2 dimensional bounding box.
    :param int im_width:
        Parent image full width in pixels.
    :param int im_height:
        Parent image full height in pixels.

    :return: If this crop specification lies fully within the given
        bounds.  Touching the edge counts as being "fully within".
    :rtype: bool
    """
    bbox_dim = bbox.ndim
    assert bbox_dim == 2, \
        "``crop_in_bounds`` requires an image-space bounding box (2 " \
        "dimensional), given bounding box with dimension {}." \
        .format(bbox_dim)

    log = logging.getLogger('.'.join((__package__, 'crop_in_bounds')))

    ul_x, ul_y = bbox.min_vertex
    lr_x, lr_y = bbox.max_vertex

    in_bounds = True
    if not ((0 <= ul_x <= im_width) and (0 <= ul_y <= im_height)):
        log.warning("Upper-left coordinate outside image bounds ([w,h] "
                    "image dimensions: {}, given upper-left: {})"
                    .format((im_width, im_height), (ul_x, ul_y)))
        in_bounds = False
    if not ((0 <= lr_x <= im_width) and (0 <= lr_y <= im_height)):
        log.warning("Lower-right coordinate outside image bounds "
                    "([w, h] image dimensions: {}, given "
                    "lower-right: {})"
                    .format((im_width, im_height), (lr_x, lr_y)))
        in_bounds = False
    if not (((lr_x - ul_x) > 0) and ((lr_y - ul_y) > 0)):
        log.warning("Pixel crop region area is zero (ul: {}, lr: {})."
                    .format((ul_x, ul_y), (lr_x, lr_y)))
        in_bounds = False
    return in_bounds



import functools
import operator

import numpy

from smqtk_core import Configurable

from typing import Tuple, Dict, Optional, Sequence, Union


class AxisAlignedBoundingBox (Configurable):
    """
    Representation of an axis-aligned rectangular region within N-dimensional
    euclidean space.

    AxisAlignedBoundingBox currently does not support the concept of an
    "empty" region since it requires a min and max vertex to be set. We
    currently do not consider a zero-area region "empty" (represents a spatial
    point).

    The class attributes ``EQUALITY_ATOL`` and ``EQUALITY_RTOL`` are used as
    the tolerance attributes when comparing equality between two
    AxisAlignedBoundingBox instances. These may be changed on the class level
    to change the desired tolerance used at runtime. This cannot be changed
    on specific instances due to the use of python slots.

    Number of vertices of a hyper-cube: ``2**D``, where ``D`` is the number of
    dimensions.
    """

    __slots__ = 'min_vertex', 'max_vertex'

    # Same tolerance values as default on ``numpy.allclose``. These may be
    # changed on the class level to affect tolerance used when comparing
    # AxisAlignedBoundingBoxes at runtime.
    EQUALITY_ATOL = 1.e-8
    EQUALITY_RTOL = 1.e-5

    def __init__(
        self, min_vertex: Sequence[Union[int, float]],
            max_vertex: Sequence[Union[int, float]]) -> None:
        """
        Create a new AxisAlignedBoundingBox from the given minimum and maximum
        euclidean-space vertex.

        :param collections.abc.Sequence[int|float] min_vertex:
            Minimum bounding vertex of the (hyper) rectangle.
        :param collections.abc.Sequence[int|float] max_vertex:
            Maximum bounding vertex of the (hyper) rectangle.

        :raises ValueError:
            When input vertices are not both 1 dimensional in
            shape, match in vertex dimensionality or if ``max_vertex`` is not
            greater-than-or-equal to ``min_vertex``.

        """
        # TODO: Default ``max_vertex`` to ``None`` to ease the creation of
        #       "points".
        self._set_vertices(min_vertex, max_vertex)

        if not (self.min_vertex.ndim == self.max_vertex.ndim == 1):
            raise ValueError("One or both vertices provided had more than "
                             "one array dimension (min_vertex.ndim == {}, "
                             "max_vertex.ndim == {})."
                             .format(self.min_vertex.ndim,
                                     self.max_vertex.ndim))
        if self.min_vertex.size != self.max_vertex.size:
            raise ValueError("Both vertices provided are not the same "
                             "dimensionality (min_vertex = {}, "
                             "max_vertex = {})."
                             .format(self.min_vertex.size,
                                     self.max_vertex.size))
        if not (self.max_vertex >= self.min_vertex).all():
            raise ValueError("The maximum vertex was not strictly >= the "
                             "minimum vertex.")

    def __str__(self) -> str:
        return "<{} [{}, {}]>"\
            .format(self.__class__.__name__, self.min_vertex, self.max_vertex)

    def __repr__(self) -> str:
        return "<{}.{} min_vertex={} max_vertex={}>"\
            .format(self.__class__.__module__, self.__class__.__name__,
                    self.min_vertex, self.max_vertex)

    def __hash__(self) -> int:
        return hash((tuple(self.min_vertex), tuple(self.max_vertex)))

    def __eq__(self, other: object) -> bool:
        """
        Two bounding boxes are equal if the describe the same spatial area.

        :param AxisAlignedBoundingBox other:
            Other bounding box instance to test equality against.

        :return: If this and `other` describe the same spatial area.
        :rtype: bool
        """
        if not isinstance(other, AxisAlignedBoundingBox):
            return False
        # Should tolerances be parameterized in constructor?
        return (numpy.allclose(self.min_vertex, other.min_vertex,
                               rtol=self.EQUALITY_RTOL,
                               atol=self.EQUALITY_ATOL) and
                numpy.allclose(self.max_vertex, other.max_vertex,
                               rtol=self.EQUALITY_RTOL,
                               atol=self.EQUALITY_ATOL))

    def __ne__(self, other: object) -> bool:
        return not (self == other)

    def __getstate__(self) -> Tuple:
        return (
            self.min_vertex.tolist(),
            self.max_vertex.tolist(),
        )

    def __setstate__(self, state: Tuple) -> None:
        self._set_vertices(*state)

    def _set_vertices(
        self, min_v: Sequence[Union[int, float]],
            max_v: Sequence[Union[int, float]]) -> None:
        self.min_vertex = numpy.asarray(min_v)
        self.min_vertex.flags.writeable = False
        self.max_vertex = numpy.asarray(max_v)
        self.max_vertex.flags.writeable = False

    def get_config(self) -> Dict:
        return {
            'min_vertex': self.min_vertex.tolist(),
            'max_vertex': self.max_vertex.tolist(),
        }

    def intersection(self, other: "AxisAlignedBoundingBox") -> Optional["AxisAlignedBoundingBox"]:
        """
        Get the AxisAlignedBoundingBox that represents the intersection between
        this box and the given ``other`` box.

        :param AxisAlignedBoundingBox other:
            An other box to get the intersection of. If there is no
            intersection ``None`` is returned.

        :return: An AxisAlignedBoundingBox instance if there is an intersection
            or None if there is no intersection.
        :rtype: AxisAlignedBoundingBox | None
        """
        inter_min_v = numpy.maximum(self.min_vertex, other.min_vertex)  # type: ignore
        inter_max_v = numpy.minimum(self.max_vertex, other.max_vertex)
        # Constructor allows zero area boxes, but that is not a valid
        # intersection. Returning None if the minimum of the difference between
        # the intersection's maximum and minimum vertices is <= 0.
        if (inter_max_v - inter_min_v).min() <= 0:
            return None
        return AxisAlignedBoundingBox(inter_min_v, inter_max_v)

    @property
    def ndim(self) -> int:
        """
        :return: The number of dimensions this bounding volume covers.
        :rtype: int
        """
        # we know because of assert in constructor that both min and max vertex
        # must match in coordinate dimensionality, so we can draw this from
        # either.
        return self.min_vertex.size

    @property
    def deltas(self) -> numpy.ndarray:
        """
        Get the lengths of this bounding box's edges along its dimensions.

        I.e. if this bounding box is 2-dimensional, this returns the [width,
        height] of the bounding box.

        :return: Array of dimension deltas.
        :rtype: numpy.ndarray[int|float]
        """
        return self.max_vertex - self.min_vertex

    @property
    def dtype(self) -> numpy.dtype:
        """
        :return: Most representative data type required to fully express this
            bounding box.
        :rtype: numpy.dtype
        """
        return self.deltas.dtype

    @property
    def hypervolume(self) -> float:
        """
        :return: The volume of this [hyper-dimensional] spatial bounding box.
            Unit of volume depends on the dimensionality of the vertices
            provided.
        :rtype: float
        """
        return functools.reduce(operator.mul,
                                self.max_vertex - self.min_vertex)



import collections
from contextlib import contextmanager
from distutils.version import LooseVersion
import tempfile
from typing import Dict, Optional, List, Sequence, Set, Union, Iterable
import warnings

import numpy as np
from six.moves import range

from smqtk_dataprovider import DataElement
from smqtk_image_io.interfaces.image_reader import ImageReader
from smqtk_image_io.utils.image import crop_in_bounds
from smqtk_image_io.bbox import AxisAlignedBoundingBox

try:
    import osgeo  # type: ignore
    import osgeo.gdal as gdal  # type: ignore
    import osgeo.gdal_array as gdal_array  # type: ignore
except ImportError:
    osgeo = gdal = gdal_array = None


###############################################################################
# Plugin helper function(s)
#
# TODO: Move appropriate functions to a ``smqtk.utils.gdal`` module.

def get_gdal_driver_supported_mimetypes() -> Set[str]:
    """
    Get a set of mimetype strings that currently available GDAL drivers
    support.

    :return: Set of MIMETYPE strings.
    """
    # look for DMD_MIMETYPE metadata key in available drivers (available in
    # both versions 1 and 2)
    try:
        return get_gdal_driver_supported_mimetypes.cache  # type: ignore
    except AttributeError:
        m_set: Set[str] = set()
        get_gdal_driver_supported_mimetypes.cache = m_set  # type: ignore
        m_key = gdal.DMD_MIMETYPE

        for i in range(gdal.GetDriverCount()):
            d = gdal.GetDriver(i)
            # A driver might be for a file type that has no associated
            # mimetype.
            d_mimetype = d.GetMetadata().get(m_key, None)
            if d_mimetype is not None:
                m_set.add(d_mimetype)

        return m_set


def get_possible_gdal_gci_values() -> Set[int]:
    """
    Get the set of possible gdal.GCI_* values.

    This function caches the constructed set as this should not change during
    the runtime of an application.

    :return: The set of possible gdal.GCI_* values.
    """
    try:
        return get_possible_gdal_gci_values.cache  # type: ignore
    except AttributeError:
        s = set(map(lambda a: getattr(gdal, a),
                    [attr for attr in dir(gdal) if attr.startswith("GCI_")]))
        get_possible_gdal_gci_values.cache = s  # type: ignore
        return s


def get_gdal_gci_abbreviation_map() -> Dict[str, int]:
    """
    Abbreviation mapping::

        'a' = gdal.GCI_AlphaBand
        'b' = gdal.GCI_BlueBand
        'g' = gdal.GCI_GreenBand
        'r' = gdal.GCI_RedBand
        'c' = gdal.GCI_CyanBand
        'm' = gdal.GCI_MagentaBand
        'y' = gdal.GCI_YellowBand
        'h' = gdal.GCI_HueBand
        's' = gdal.GCI_SaturationBand
        'l' = gdal.GCI_LightnessBand

    This function caches the constructed set as this should not change during
    the runtime of an application.

    :return: the new or cached character-to-GCI value map for channels for which
        a single letter abbreviation makes sense.  Character keys are in lower
        case.
    :rtype: dict[str, int]
    """
    try:
        return get_gdal_gci_abbreviation_map.map_cache  # type: ignore
    except AttributeError:
        m = {
            'a': gdal.GCI_AlphaBand,
            'b': gdal.GCI_BlueBand,
            'g': gdal.GCI_GreenBand,
            'r': gdal.GCI_RedBand,
            'c': gdal.GCI_CyanBand,
            'm': gdal.GCI_MagentaBand,
            'y': gdal.GCI_YellowBand,
            'h': gdal.GCI_HueBand,
            's': gdal.GCI_SaturationBand,
            'l': gdal.GCI_LightnessBand
        }
        get_gdal_gci_abbreviation_map.map_cache = m  # type: ignore
        return m


def map_gci_list_to_names(gci_list: Iterable[int]) -> List[str]:
    """
    Translate a sequence of GDAL GCI values into a list of their string names.

    Pre-condition: Integers provided are valid color interpretation constants.

    :param collections.abc.Iterable[int] gci_list:
        Integer GDAL color interpretation integer constants sequence.
    :return: List of o the string names for each color interpretation constant
        provided
    :rtype: list[str]
    """
    return [gdal.GetColorInterpretationName(gci_int) for gci_int in gci_list]


###############################################################################
# Dataset load context managers
#
# These functions should be decorated with ``@contextmanager``, take in a
# single positional argument that is a SMQTK DataElement instance, and yield a
# GDAL Dataset instance. Functions should clean-up after themselves in their
# ``finally`` clause.
#

@contextmanager
def load_dataset_tempfile(data_element: DataElement) -> "gdal.Dataset":
    """
    Load GDAL Dataset from element by first writing it to a temporary file.

    :param smqtk.representation.DataElement data_element:
        Element to load dataset from.

    :return: GDAL Dataset
    :rtype: gdal.Dataset

    """
    fp = data_element.write_temp()
    try:
        yield gdal.Open(fp)
    finally:
        data_element.clean_temp()


@contextmanager
def load_dataset_vsimem(data_element: DataElement) -> "gdal.Dataset":
    """
    Load GDAL dataset from element by writing its bytes to a virtual file
    and loading a dataset from that virtual file.

    Requires GDAL major version 2 or greater.

    :param smqtk.representation.DataElement data_element:
        Element to load dataset from.

    :return: GDAL Dataset
    :rtype: gdal.Dataset

    """
    # Unguarded next() call is OK in this case because the generator returned
    # by ``_get_candidate_names()`` does not terminate.
    # noinspection PyProtectedMember
    get_candidate_names = tempfile._get_candidate_names  # type: ignore
    tmp_vsimem_path = '/vsimem/{}'.format(
        next(get_candidate_names())  # lgtm[py/unguarded-next-in-generator]
    )
    gdal.FileFromMemBuffer(tmp_vsimem_path, data_element.get_bytes())
    try:
        yield gdal.Open(tmp_vsimem_path)
    finally:
        rc = gdal.Unlink(tmp_vsimem_path)
        if rc != 0:
            raise RuntimeError("Failed to gdal.Unlink virtual file '{}' "
                               "containing bytes from {}."
                               .format(tmp_vsimem_path, data_element))


###############################################################################
# Base GDAL reader classes


class GdalImageReader (ImageReader):
    """
    Use GDAL to read raster image pixel data and returns an image matrix in
    the [height, width, channel] format for multi-channel imagery and just
    [height, width] for single channel imagery.
    """

    LOAD_METHOD_TEMPFILE = 'tempfile'
    LOAD_METHOD_VSIMEM = 'vsimem'
    LOAD_METHOD_CONTEXTMANAGERS = {
        LOAD_METHOD_TEMPFILE: load_dataset_tempfile,
        LOAD_METHOD_VSIMEM: load_dataset_vsimem,
    }

    @classmethod
    def is_usable(cls) -> bool:
        """
        Check whether this class is available for use.

        Since certain plugin implementations may require additional
        dependencies that may not yet be available on the system,
        this method should check for those dependencies and return a
        boolean saying if the implementation is usable.

        NOTES:
            - This should be a class method
            - When an implementation is deemed not usable, this should
              emit a warning detailing why the implementation is not
              available for use.

        :return: Boolean determination of whether this implementation is
            usable.
        :rtype: bool

        """
        if osgeo is None:
            warnings.warn("GdalImageReader implementation not available "
                          "because ``osgeo`` is not importable.")
            return False
        return True

    def __init__(self, load_method: str = LOAD_METHOD_TEMPFILE,
                 channel_order: Optional[Union[str, Sequence[int]]] = None) -> None:
        """
        Use GDAL to read raster image pixel data and returns an image matrix in
        the format native to the input data.

        Channel Order
        -------------
        A custom selection of input image channels as well as their order may
        be specified by the ``channel_order`` parameter.  This may either be a
        string that uses channel abbreviations or a sequence of integers
        specifying GDAL GCI constants (``osgeo.gdal.GCI_*``).  If a string is
        provided, we cast to lower case as a standardization.

        If the source image does not report color interpretations for its bands
        then any specification of ``channel_order`` will result in an error when
        ``load_as_matrix`` is called (unable to map requested channels to bands
        reporting as type 0, or "unknown").

        See :func:`get_gdal_gci_abbreviation_map` for supported abbreviations.

        Load methods
        ------------

        ``tempfile``
        ^^^^^^^^^^^^
        Loader that writes the DataElement's bytes to a temporary file on
        disk first and then "opens" that file as a GDAL Dataset. This
        method performs double the I/O operations (write to disk and
        then immediately read from it again) but uses less RAM (only
        one "copy" of the image is loaded at a time) and is universally
        supported across GDAL versions.

        ``vsimem``
        ^^^^^^^^^^
        This method is more efficient in regards to run-time, however two
        copies of the image are loaded into RAM (the bytes of image for the
        virtual file and then image matrix itself). This method is also only
        functional in GDAL version 2 and above currently. A RuntimeError is
        raised if the currently imported GDAL is not version 2 or greater.

        :param load_method:
            Method of loading GDAL Dataset from a DataElement.  This must be
            one of the ``GdalImageReader.LOAD_METHOD_*`` options.
        :param channel_order:
            Optional specific selection and order of channels read from the
            image to be included in the output matrix.  See above for more
            details.

        :raises RuntimeError: The ``vsimem`` load method was specified but the
            currently imported GDAL version is not >= 2.

        """
        super(GdalImageReader, self).__init__()

        self._channel_order = channel_order
        # The channel order in known GDAL GCI integer values. This is None when
        # channel_order is none, otherwise is a sequence of valid GDAL GCI
        # integer values.
        self._channel_order_gci: Optional[List[int]] = None
        if channel_order is not None:
            # Is Sequence check.
            if not isinstance(channel_order, collections.abc.Sequence):
                raise ValueError("Channel order must be a sequence in order "
                                 "to discern order! Given type: {}"
                                 .format(type(channel_order)))
            # Cannot be given an empty order sequence.
            if len(channel_order) == 0:
                raise ValueError("Invalid channel order, must request at "
                                 "least one band. Given: '{}'"
                                 .format(channel_order))
            # If using an abbreviation string, make sure all characters match a
            # known abbreviation.
            if isinstance(channel_order, str):
                channel_order = channel_order.lower()
                self._channel_order = channel_order
                abb_map = get_gdal_gci_abbreviation_map()
                # Set will be non-empty if there is an invalid character.
                valid_chr_set = set(abb_map)
                # noinspection PyTypeChecker
                str_diff_set = set(channel_order).difference(valid_chr_set)
                if str_diff_set:
                    raise ValueError("Invalid abbreviation character in given "
                                     "channel order. Invalid characters: {}. "
                                     "Valid characters: {}."
                                     .format(list(str_diff_set),
                                             list(valid_chr_set)))
                # Cache channel order in GCI translated values
                self._channel_order_gci = [abb_map[a] for a in channel_order]
            # When not string, make sure integer values are in the reported set
            # values from GDAL, otherwise we'll get a runtime error later.
            else:
                valid_int_set = set(get_possible_gdal_gci_values())
                int_diff_set = set(channel_order).difference(valid_int_set)
                if int_diff_set:
                    raise ValueError("Invalid GDAL band integer values given. "
                                     "Given invalid values: {}."
                                     .format(list(int_diff_set)))
                # Given channel order was already in GCI values, mirror in
                # expected attribute.
                self._channel_order_gci = list(channel_order)

        self._load_method = load_method.lower()
        # 1. Check that the given load method is one that we support.
        # 2. Check that GDAL version >= 2 if VSIMEM method specified
        if self._load_method not in self.LOAD_METHOD_CONTEXTMANAGERS:
            raise ValueError("Load method provided not a valid value (given "
                             "'{}'). Must be one of: {}."
                             .format(load_method,
                                     self.LOAD_METHOD_CONTEXTMANAGERS))
        elif self._load_method == self.LOAD_METHOD_VSIMEM:
            gdal_major_version: int = int(
                LooseVersion(osgeo.__version__).version[0]
            )
            if gdal_major_version < 2:
                raise RuntimeError("Load method '{}' was specified, "
                                   "but required GDAL version of 2 is not "
                                   "satisfied (imported version: {})."
                                   .format(self._load_method,
                                           osgeo.__version__))

    def get_config(self) -> Dict:
        """
        Return a JSON-compliant dictionary that could be passed to this class's
        ``from_config`` method to produce an instance with identical
        configuration.

        In the common case, this involves naming the keys of the dictionary
        based on the initialization argument names as if it were to be passed
        to the constructor via dictionary expansion.

        :return: JSON type compliant configuration dictionary.
        :rtype: dict

        """
        return {
            'load_method': self._load_method,
            'channel_order': self._channel_order,
        }

    def valid_content_types(self) -> Set[str]:
        """
        :return: A set valid MIME types that are "valid" within the implementing
            class' context.
        :rtype: set[str]
        """
        return get_gdal_driver_supported_mimetypes()

    def _load_as_matrix(
        self, data_element: DataElement,
            pixel_crop: AxisAlignedBoundingBox = None) -> np.ndarray:
        """
        Internal method to be implemented that attempts loading an image
        from the given data element and returning it as an image matrix.

        Pre-conditions:
            - ``pixel_crop`` has a non-zero volume and is composed of integer
              types.

        :param smqtk.representation.DataElement data_element:
            DataElement to load image data from.
        :param None|smqtk.representation.AxisAlignedBoundingBox pixel_crop:
            Optional pixel crop region to load from the given data.  If this
            is provided it must represent a valid sub-region within the loaded
            image, otherwise a RuntimeError is raised.

        :raises RuntimeError: A crop region was specified but did not specify a
            valid sub-region of the image.

        :return: Numpy ndarray of the image data. Specific return format is
            implementation dependant.
        :rtype: np.ndarray

        """
        if data_element.is_empty():
            raise ValueError(
                "GdalImageReader cannot load 0-sized data (no bytes in {})."
                .format(data_element))
        load_cm = self.LOAD_METHOD_CONTEXTMANAGERS[self._load_method]
        with load_cm(data_element) as gdal_ds:  # type: gdal.Dataset
            img_width = gdal_ds.RasterXSize
            img_height = gdal_ds.RasterYSize

            # GDAL wants [x, y, width, height] as the first 4 positional
            # arguments to ``ReadAsArray``.
            xywh = [0, 0, img_width, img_height]
            if pixel_crop:
                if not crop_in_bounds(pixel_crop, img_width, img_height):
                    raise RuntimeError("Crop provided not within input image. "
                                       "Image shape: {}, crop: {}"
                                       .format((img_width, img_height),
                                               pixel_crop))
                # This is testing faster than ``np.concatenate``.
                xywh = \
                    pixel_crop.min_vertex.tolist() + pixel_crop.deltas.tolist()

            # Select specific channels if they are present in this dataset, or
            # just get all of them
            if self._channel_order is not None:
                assert self._channel_order_gci is not None, (
                    "When a channel-order is set, the GCI equivalent should "
                    "also be set."
                )
                # Map raster bands from CI value to band index.
                # - GDAL uses 1-based indexing.
                band_ci_to_idx = {
                    gdal_ds.GetRasterBand(b_i).GetColorInterpretation(): b_i
                    for b_i in range(1, gdal_ds.RasterCount+1)
                }
                gci_diff = (
                    set(self._channel_order_gci).difference(band_ci_to_idx)
                )
                if gci_diff:
                    raise RuntimeError(
                        "Data element did not provide channels required to "
                        "satisfy requested channel order {}.  "
                        "Data had bands: {} (missing {})."
                        .format(map_gci_list_to_names(self._channel_order_gci),
                                map_gci_list_to_names(band_ci_to_idx),
                                map_gci_list_to_names(gci_diff)))
                # Initialize a matrix to read band image data into
                # TODO: Handle when there are no bands?
                band_dtype = gdal_array.GDALTypeCodeToNumericTypeCode(
                    gdal_ds.GetRasterBand(1).DataType
                )
                if len(self._channel_order_gci) > 1:
                    img_mat = np.ndarray([xywh[3], xywh[2],
                                          len(self._channel_order_gci)],
                                         dtype=band_dtype)
                    for i, gci in enumerate(self._channel_order_gci):
                        #: :type: gdal.Band
                        b = gdal_ds.GetRasterBand(band_ci_to_idx[gci])
                        b.ReadAsArray(*xywh, buf_obj=img_mat[:, :, i])
                else:
                    img_mat = np.ndarray([xywh[3], xywh[2]],
                                         dtype=band_dtype)
                    gci = self._channel_order_gci[0]
                    b = gdal_ds.GetRasterBand(band_ci_to_idx[gci])
                    b.ReadAsArray(*xywh, buf_obj=img_mat)
            else:
                img_mat = gdal_ds.ReadAsArray(*xywh)
                if img_mat.ndim > 2:
                    # Transpose into [height, width, channel] format.
                    img_mat = img_mat.transpose(1, 2, 0)

        return img_mat



import abc

import numpy

from smqtk_core import Plugfigurable
from smqtk_dataprovider import ContentTypeValidator, DataElement
from smqtk_image_io.bbox import AxisAlignedBoundingBox
from typing import Optional


class ImageReader (Plugfigurable, ContentTypeValidator):
    """
    Interface for algorithms that load a raster image matrix from a data
    element.
    """

    __slots__ = ()

    @staticmethod
    def _get_matrix_property(data_element: DataElement) -> Optional[numpy.ndarray]:
        """
        Central method of getting and checking the matrix property of a
        DataElement.

        :param smqtk.representation.DataElement data_element:
            SMQTK DataElement instance.

        :raises AttributeError: If the element given does not have a ``matrix``
            attribute.
        :raises AssertionError: If the element's matrix property does not match
            a None or ndarray value type.

        :return: Matrix property value if it is None or an ndarray.
        """

        mat_prop = data_element.matrix  # type: ignore
        assert mat_prop is None or isinstance(mat_prop, numpy.ndarray), \
            "Element `matrix` property return should either be a matrix " \
            "or None. Got {} instead.".format(type(mat_prop))
        return mat_prop

    def is_valid_element(self, data_element: DataElement) -> bool:
        """
        Check if the given DataElement instance reports a content type that
        matches one of the MIME types reported by ``valid_content_types``.

        This override checks if the ``DataElement`` has the ``matrix`` property
        as the ``MatrixDataElement`` would provide, and that its value of an
        expected type.

        :param smqtk.representation.DataElement data_element:
             Data element instance to check.

        :return: True if the given element has a valid content type as reported
            by ``valid_content_types``, and False if not.
        :rtype: bool
        """
        try:
            # If the given data element looks like a MatrixDataElement
            self._get_matrix_property(data_element)
            return True
        except AttributeError:
            # Otherwise proceed through traditional route.
            return super(ImageReader, self)\
                .is_valid_element(data_element)

    def load_as_matrix(
        self, data_element: DataElement,
            pixel_crop: Optional[AxisAlignedBoundingBox] = None) \
            -> Optional[numpy.ndarray]:
        """
        Load an image matrix from the given data element.

        **Matrix Property Shortcut.**
        If the given DataElement instance defines a ``matrix`` property this
        method simply returns that.  This is intended to interface with
        instances of
        :py:class:`smqtk.representation.data_element.matrix.MatrixDataElement`.

        **Loading From Bytes.**
        When not loading from a short-cut matrix, matrix return format is
        ``ImageReader`` implementation dependant. Implementations of this
        interface should specify and describe their return type.

        Aside from the exceptions documented below, other exceptions may be
        raised when an image fails to load that are implementation dependent.

        :param smqtk.representation.DataElement data_element:
            DataElement to load image data from.
        :param None|smqtk.representation.AxisAlignedBoundingBox pixel_crop:
            Optional bounding box specifying a pixel sub-region to load from the
            given data.  If this is provided it must represent a valid
            sub-region within the loaded image, otherwise a RuntimeError is
            raised.  Handling of non-integer aligned boxes are implementation
            dependant.

        :raises RuntimeError: A crop region was specified but did not specify a
            valid sub-region of the image.
        :raises AssertionError: The ``data_element`` provided defined a
            ``matrix`` attribute/property, but its access did not result in an
            expected value.
        :raises ValueError:
            This error is raised when:
                - The given ``data_element`` was not of a valid content type.
                - A ``pixel_crop`` bounding box was provided but was zero
                  volume.
                - ``pixel_crop`` bounding box vertices are not fully
                  represented by integers.

        :return: Numpy ndarray of the image data. Specific return format is
            implementation dependant.
        :rtype: numpy.ndarray

        """
        if pixel_crop:
            if pixel_crop.hypervolume == 0:
                raise ValueError("Volume of crop bounding box must be greater "
                                 "than 0. Given: {}".format(pixel_crop))
            elif not issubclass(pixel_crop.dtype.type, numpy.integer):
                raise ValueError("Crop bounding box must be composed of "
                                 "integer coordinates. Given bounding box "
                                 "with dtype {}.".format(pixel_crop.dtype.type))

        try:
            # If the given data element looks like a MatrixDataElement, simply
            # return the stored matrix property.
            return self._get_matrix_property(data_element)
        except AttributeError:
            # Any other data element type, attempt loading via plugin
            # implementation.
            self.raise_valid_element(data_element)
            return self._load_as_matrix(data_element, pixel_crop=pixel_crop)

    @abc.abstractmethod
    def _load_as_matrix(
        self, data_element: DataElement,
            pixel_crop: Optional[AxisAlignedBoundingBox] = None) \
            -> numpy.ndarray:
        """
        Internal method to be implemented that attempts loading an image
        from the given data element and returning it as an image matrix.

        Pre-conditions:
            - ``pixel_crop`` has a non-zero volume and is composed of integer
              types.

        :param smqtk.representation.DataElement data_element:
            DataElement to load image data from.
        :param None|smqtk.representation.AxisAlignedBoundingBox pixel_crop:
            Optional pixel crop region to load from the given data.  If this
            is provided it must represent a valid sub-region within the loaded
            image, otherwise a RuntimeError is raised.

        :raises RuntimeError: A crop region was specified but did not specify a
            valid sub-region of the image.

        :return: Numpy ndarray of the image data. Specific return format is
            implementation dependant.
        :rtype: numpy.ndarray

        """


# TODO: ImageWriter interface



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['smqtk_image_io',
 'smqtk_image_io.impls',
 'smqtk_image_io.impls.image_reader',
 'smqtk_image_io.interfaces',
 'smqtk_image_io.utils']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.19.5,<2.0.0',
 'pillow>=8.2.0,<9.0.0',
 'smqtk-core>=0.18.0',
 'smqtk-dataprovider>=0.16.0']

setup_kwargs = {
    'name': 'smqtk-image-io',
    'version': '0.16.0',
    'description': 'SMQTK Image IO',
    'long_description': '# SMQTK - Image-IO\n\n## Intent\nThis package is intended to provide the interfaces and tools for working with image input.\n\nSpecifically, there are two readers used to interact with images. One uses GDAL and the other uses PIL\n\n## Documentation\n\nYou can build the sphinx documentation locally for the most\nup-to-date reference (see also: [Building the Documentation](\nhttps://smqtk.readthedocs.io/en/latest/installation.html#building-the-documentation)):\n```bash\n# Install dependencies\npoetry install\n# Navigate to the documentation root.\ncd docs\n# Build the docs.\npoetry run make html\n# Open in your favorite browser!\nfirefox _build/html/index.html\n```\n',
    'author': 'Kitware, Inc.',
    'author_email': 'smqtk-developers@kitware.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
